'''
WormbaseREST REST API for http://rest.wormbase.org/index.html
'''
import time
import json
import urllib.request
import logging
import logging.config

import aiohttp
import asyncio
import concurrent.futures
import multiprocessing
import numpy as np

from . import load_wormbase_api_json

try:
    logging.config.fileConfig('logging.config')
except Exception:
    pass

# Create a logger object
logger = logging.getLogger(__name__)

class WormbaseAPI:

    def __init__(self,call_type, call_class, data_request):
        self.base_url_str = "https://wormbase.org/rest"
        self.max_retries = 3
        self.call_type = call_type
        self.call_class = call_class
        self.data_request = data_request
        self.wormbase_api_json = load_wormbase_api_json(call_type, call_class)
        if data_request not in self.wormbase_api_json:
            logger.error(f"No wormbase config for {data_request=}")
            self.results_doc_definition = {}
        else:
            self.results_doc_definition = self.wormbase_api_json[data_request]
        

    def _rest_api_call(self, object_id):
        url_str = f"{self.base_url_str}/{self.call_type}/{self.call_class}/{object_id}/{self.data_request}"
        logger.debug(url_str)
        retry = 0
        done = False

        api_result = None
        api_error = None

        def handle_error(error_msg):
            print(error_msg)
            logger.debug(error_msg)
            nonlocal done, retry, api_error
            retry +=1
            if retry >= self.max_retries:
                done = True
                api_error = error_msg

        while not done:
            try:
                url = urllib.request.urlopen(url_str)
                if url.getcode() == 200:
                    done = True
                    response_text = url.read().decode('utf-8')
                    api_result = json.loads(response_text)
                elif url.getcode() == 429:
                    handle_error(f"Request limiter hit. waiting 2 seconds [Retry: {retry + 1}] code: {url.getcode()}")
                    time.sleep(2)
                else:
                    handle_error(f"Failed to retrieve data. | Retry- {retry +1} | Response code- {url.getcode()}")
            except Exception as ex:
                aviod_logging_interpolation=f"Error while calling {url_str} | {str(ex)}"
                logger.error(aviod_logging_interpolation)
                error_msg=f"Check if you have a connection!! | Retry- {retry+1} | Response msg- {str(ex)}"
                time.sleep(3)
                if isinstance(ex, urllib.error.HTTPError):
                    if ex.code == 500:
                        error_msg=f"Check the format of the http request [Retry: {retry + 1}]\nurl:{url_str}\ncode: {str(ex)}"
                else:
                    error_msg=f"Check if you have a connection!! | Retry- {retry+1} | Response msg- {str(ex)}"
                handle_error(error_msg)

        if api_result is None:
            api_result = {"rest_api_error": api_error}
        
        if logger.isEnabledFor(logging.DEBUG):
            pretty_data = json.dumps(api_result, indent=4)
            with open('http_response.json', 'w') as file:
                file.write(pretty_data)
                
        return api_result

    async def _async_rest_api_call(self, object_id):
            url_str = f"{self.base_url_str}/{self.call_type}/{self.call_class}/{object_id}/{self.data_request}"
            retry = 0
            done = False
            api_result = None
            api_error = None

            def handle_error(error_msg):
                print(error_msg)
                logger.debug(error_msg)
                nonlocal done, retry, api_error
                retry +=1
                if retry >= self.max_retries:
                    done = True
                    api_error = error_msg


            async with aiohttp.ClientSession() as session:
                while not done:
                    try:
                        async with session.get(url_str) as response:
                            if response.status == 200:
                                done = True
                                response_text = await response.text()
                                api_result =  json.loads(response_text)
                            elif response.status == 429:
                                error_msg = f"Request limiter hit. Waiting 2 seconds [Retry: {retry + 1}]"
                                handle_error(error_msg)
                                await asyncio.sleep(2)
                            else:
                                error_msg = f"Failed to retrieve data. Retry- {retry + 1}, Code- {response.status}"
                                handle_error(error_msg)
                    except Exception as ex:
                        logger.error(f"Error calling {url_str}: {str(ex)}")
                        await asyncio.sleep(3)
                        error_msg = f"Error on retry {retry + 1}: {str(ex)}"
                        handle_error(error_msg)

            return api_result if api_result else {"error": api_error}
        
    
    def _get_json_element(self, json_data, path):
        result = json_data
        try:
            for key in path:
                result = result[key]
        except Exception: #KeyError TypeError
            result = None
        return result


    def _extract_empty_dict(self, json_obj):
        if isinstance(json_obj, dict):
            return {k: self._extract_empty_dict(v) for k, v in json_obj.items() if v and self._extract_empty_dict(v)}
        elif isinstance(json_obj, list):
            return [self._extract_empty_dict(v) for v in json_obj if v and self._extract_empty_dict(v)]
        else:
            return json_obj
    
    def _extract_single_element_lists(self, json_obj):
        if isinstance(json_obj, dict):
            for key, value in json_obj.items():
                json_obj[key] = self._extract_single_element_lists(value)
        elif isinstance(json_obj, list):
            # If list length is 1 remove list
            if len(json_obj) == 1:
                return self._extract_single_element_lists(json_obj[0])
            else:
                return [self._extract_single_element_lists(item) for item in json_obj]
        return json_obj

    def _extract_skip_elements(self, json_obj):
        if isinstance(json_obj, dict):
            for key, value in json_obj.items():
                if key == "SKIP":
                    json_obj = self._extract_skip_elements(value)
                else:
                    json_obj[key] = self._extract_skip_elements(value)
        elif isinstance(json_obj, list):
            return [self._extract_skip_elements(item) for item in json_obj]
        return json_obj
    
    def _parse_data(self, data_to_process, doc_definition, results_dict):
        # data_request_item_nm="description" data_request_item=["fields", "concise_description","data","text"]
        # data_request_item_nm="author"      data_request_item={ "ROOT": ["author"], "CONCAT": ["label"] }
        # Just used to shorten the call length to make code more readable
        get_json = self._get_json_element
        for data_request_item_nm, data_request_item in doc_definition.items():
            #logger.debug(f"{data_request_item_nm=}{data_request_item=}")
            if isinstance(data_request_item, list):
                data_request_item_path = data_request_item
                widget_item = get_json(data_to_process, data_request_item_path)
                if widget_item is not None:
                    results_dict[data_request_item_nm] = widget_item
            elif isinstance(data_request_item, dict):
                #logger.debug(f"BEFORE {data_request_item=}, {data_request_item['ROOT']=}")
                #pretty_data = json.dumps(data_to_process, indent=4)
                #logger.debug(f"BEFORE {pretty_data}")
                sub_data_to_process = get_json(data_to_process, data_request_item["ROOT"])

                if sub_data_to_process is not None:
                    #pretty_data = json.dumps(sub_data_to_process, indent=4)
                    #logger.debug(f"AFTER {pretty_data}")
                    if "CONCAT" in data_request_item:
                        sub_results_str = ""
                        for sub_data_item in sub_data_to_process:
                            sub_results = self._parse_data(sub_data_item, data_request_item, {})
                            if "CONCAT" in sub_results:
                                sub_results_str +=str(f"{sub_results['CONCAT']}|")
                        results_dict[data_request_item_nm] = sub_results_str[:-1]
                        
                    else:
                        if isinstance(sub_data_to_process, dict):
                            results_dict[data_request_item_nm] = self._parse_data(sub_data_to_process, data_request_item, {})
                        else: #It is a list
                            sub_results_list = []
                            for sub_data_item in sub_data_to_process:
                                list_item_to_append = self._parse_data(sub_data_item, data_request_item, {})
                                sub_results_list.append(list_item_to_append)
                            results_dict[data_request_item_nm] = sub_results_list

            else:
                logger.err("parse_data() ERROR Did not expect to get here!!")

        # Post processing
        results_dict = self._extract_empty_dict(results_dict)
        results_dict = self._extract_skip_elements(results_dict)
        results_dict = self._extract_single_element_lists(results_dict)
        return results_dict


    def get_wormbase_data(self, object_id, map_result=True):
        if object_id is None:
            raise Exception("objectID cannot be null!")
    
        rest_api_call_results = self._rest_api_call(object_id)
        if "rest_api_error" in rest_api_call_results:
            return rest_api_call_results
        
        if not map_result:
            return rest_api_call_results

        ret_dict = {}
        ret_dict = self._parse_data(rest_api_call_results, self.results_doc_definition, ret_dict)
        return ret_dict

    async def async_get_wormbase_data(self, object_id, map_result=True):

        if object_id is None:
            raise Exception("objectID cannot be null!")
    
        rest_api_call_results = await self._async_rest_api_call(object_id)
        if "rest_api_error" in rest_api_call_results:
            return rest_api_call_results
        
        if not map_result:
            return rest_api_call_results

        ret_dict = {}
        ret_dict = self._parse_data(rest_api_call_results, self.results_doc_definition, ret_dict)
        return ret_dict


    def get_wormbase_data_lst(self, object_id_list):
        ret_lst = []
        for object_id in object_id_list:
            ret_val = self.get_wormbase_data(object_id)
            ret_lst.append({object_id:ret_val})
        return ret_lst

    def get_wormbase_data_cpu(self, object_id_list, max_workers=2):
        max_workers = min(max_workers, multiprocessing.cpu_count())
        sub_lists = np.array_split(object_id_list, max_workers)

        with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers) as executor:
            futures = [executor.submit(self.get_wormbase_data_lst, args) for args in sub_lists]
            results = [future.result() for future in futures]
        
        flattened_results = [item for sublist in results for item in sublist]
        return flattened_results




import importlib.util
import os
import pandas as pd

module_name = __name__
module_spec = importlib.util.find_spec(module_name)
package_path = os.path.dirname(module_spec.origin)
file_path = os.path.join(package_path, 'data', '2022_JCR_IF.csv')

impact_factor_df = pd.read_csv(file_path)



import pytest
import json
import inspect
import os
from pub_worm.wormbase.wormbase_api import WormbaseAPI

DUMP_API_CALL = True
def dump_api_call(function_name, actual_result):
    if DUMP_API_CALL:
        pretty_data = json.dumps(actual_result, indent=4)
        output_path = "output"
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        with open(f"{output_path}/{function_name}.json", 'w') as file:
                file.write(pretty_data)
        print(pretty_data)

def test_from_class_gene_request_gene_ontology_data():
    function_name = inspect.currentframe().f_code.co_name
    wormbase_id = "WBGene00195248"
    wormbase_api = WormbaseAPI("field", "gene", "gene_ontology_summary")
    actual_result = wormbase_api.get_wormbase_data(wormbase_id)
    dump_api_call(function_name, actual_result)
    
    expected_result = "GO:0022890"
    assert actual_result['gene_ontology_summary']['Molecular_function']['go_id'] == expected_result

def test_from_class_gene_request_references():
    function_name = inspect.currentframe().f_code.co_name
    wormbase_id = "WBGene00008205"
    #wormbase_id = "WBGene00195248"
    wormbase_api = WormbaseAPI("field", "gene", "references")
    actual_result = wormbase_api.get_wormbase_data(wormbase_id)
    dump_api_call(function_name, actual_result)
    
    expected_result = "WBPaper00043057"
    assert actual_result['references_list'][0]['wbp_id'] == expected_result

def test_from_class_gene_request_overview():
    function_name = inspect.currentframe().f_code.co_name
    wormbase_id = "WBGene00008205"
    #wormbase_id = "WBGene00195248"
    wormbase_api = WormbaseAPI("widget", "gene", "overview")
    actual_result = wormbase_api.get_wormbase_data(wormbase_id)
    dump_api_call(function_name, actual_result)
    
    expected_result = "sams-1"
    assert actual_result['wb_gene_name'] == expected_result

def test_from_class_paper_request_pmid():
    function_name = inspect.currentframe().f_code.co_name
    wormbase_paper_id = "WBPaper00027772"
    wormbase_api = WormbaseAPI("field", "paper", "pmid")
    actual_result = wormbase_api.get_wormbase_data(wormbase_paper_id)
    dump_api_call(function_name, actual_result)
    
    expected_result = "16291722"
    assert actual_result['pm_id'] == expected_result

def test_from_class_paper_request_abstract():
    function_name = inspect.currentframe().f_code.co_name
    wormbase_paper_id = "WBPaper00027772"
    wormbase_api = WormbaseAPI("field", "paper", "abstract")
    actual_result = wormbase_api.get_wormbase_data(wormbase_paper_id)
    dump_api_call(function_name, actual_result)
    
    expected_result = "Nephronophthisis (NPH) is a cystic kidney disorder"
    assert actual_result['wbp_abstract'].startswith(expected_result)

if __name__ == "__main__":
    pytest.main([__file__])



'''
ORCID API https://pub.orcid.org/v3.0
'''
import os
import time
import urllib.parse
import logging
import logging.config
import pandas as pd
import json
import requests

try:
    logging.config.fileConfig('logging.config')
except Exception:
    pass

# Create a logger object
logger = logging.getLogger(__name__)

class OrcidAPI:

    def __init__(self):
        self.base_url_str = "https://pub.orcid.org/v3.0"
        self.max_retries = 3
        self.timeout = 20
        self.api_key = os.environ.get('ORCID_ACCESS_TOKEN', None)


    def _rest_api_call(self, params):
        api_result = None
        api_error = "No Error Set"

        if 'orcid_id' not in params:
             logger.error("'orcid_id' param required but not passed.")
             return api_result
             
        if self.api_key is None:
             logger.error("'ORCID_ACCESS_TOKEN' is required but not found in env.")
             return api_result

        url_str = f"{self.base_url_str}/{params['orcid_id']}"

        headers = {
            "Accept": "application/vnd.orcid+json",
            "Authorization type and Access token": f"Bearer {self.api_key}"
        }

        # Query String may be used in the future
        # query = '&'.join([f"{urllib.parse.quote(k, 'utf-8')}={urllib.parse.quote(v, 'utf-8')}" for k, v in params.items()])
        # url_str = f"{url_str}?{query}"
        logger.debug(url_str)

        #self.max_retries = 3
        retry = 0
        done = False

        def handle_error(error_msg):
            print(error_msg)
            logger.debug(error_msg)
            nonlocal done, retry, api_error
            retry +=1
            if retry >= self.max_retries:
                done = True
                api_error = error_msg

        while not done:
            try:
                response = requests.get(url_str, headers=headers, timeout=self.timeout)
                response.raise_for_status()  # Check for HTTP errors
                if response.status_code == 200:
                    done = True
                    api_result = response.json()
                elif response.status_code == 429:
                    handle_error(f"Request limiter hit. waiting 2 seconds [Retry: {retry + 1}] code: {response.status_code}")
                    time.sleep(2)
                else:
                    handle_error(f"Failed to retrieve data. | Retry- {retry +1} | Response code- {response.status_code}")

            except Exception as ex:
                aviod_logging_interpolation=f"Error while calling url_str {str(ex)}"
                logger.error(aviod_logging_interpolation)
                error_msg=f"Unexpected Error | Retry- {retry+1} | Response msg- {str(ex)}"
                if isinstance(ex, requests.exceptions.HTTPError):
                        error_msg = f"Check the format of the http request [Retry: {retry + 1}] code: {str(ex)}"
                elif isinstance(ex, requests.exceptions.ConnectionError):
                        error_msg = f"Connection Error [Retry: {retry + 1}] code: {str(ex)}"
                elif isinstance(ex, requests.exceptions.Timeout):
                        error_msg = f"Timeout Error [Retry: {retry + 1}] code: {str(ex)}"
                handle_error(error_msg)

        if api_result is None:
            api_result = {"rest_api_error":api_error}

        if logger.isEnabledFor(logging.DEBUG):
            pretty_data = json.dumps(api_result, indent=4)
            with open('http_response.json', 'w', encoding="utf-8") as file:
                file.write(pretty_data)
            logger.debug(pretty_data)

        return api_result
    
    def get_orcid_data(self, orcid_id, map_result_to=None):

        if orcid_id is None:
            raise Exception("orcid_id cannot be null!")
    
        params = {'orcid_id':orcid_id}
        rest_api_call_results = self._rest_api_call(params)
        if "rest_api_error" in rest_api_call_results:
            return rest_api_call_results
        
        if map_result_to is None:
            return rest_api_call_results

        # Future code can be used to map results by parsing the json return object
        # for not we just return the whole thing!
        return rest_api_call_results




from pub_worm.impact_factor import impact_factor_df

def get_impact_factor(issn):
    # Look for ISSN or eISSN in the provided DataFrame
    match = impact_factor_df[(impact_factor_df['ISSN'] == issn) | (impact_factor_df['eISSN'] == issn)]
    if not match.empty:
        return match.iloc[0]['2021 JIF']
    else:
        return None


'''
NCBI REST API for https://eutils.ncbi.nlm.nih.gov/entrez/eutils
'''
import os
import time
import urllib.parse
import logging
import logging.config
import pandas as pd
from bs4 import BeautifulSoup
import requests
from pub_worm.impact_factor.impact_factor_lookup import get_impact_factor

try:
    logging.config.fileConfig('logging.config')
except Exception:
    pass

# Create a logger object
logger = logging.getLogger(__name__)

class EntrezAPI:

    def __init__(self):
        self.base_url_str = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils"
        self.max_retries = 3
        self.timeout = 20
        self.function = "esearch"
        self.api_key = os.environ.get('NCBI_API_KEY', None)


    def _rest_api_call(self, params, data=None):
        url_str = f"{self.base_url_str}/{self.function}.fcgi"
        params['retmode']='xml'

        if self.api_key:
            params['api_key'] = self.api_key

        query = '&'.join([f"{urllib.parse.quote(k, 'utf-8')}={urllib.parse.quote(v, 'utf-8')}" for k, v in params.items()])
        url_str = f"{url_str}?{query}"

        #self.max_retries = 3
        retry = 0
        done = False

        api_result = None
        api_error = "No Error Set"

        def handle_error(error_msg):
            print(error_msg)
            logger.debug(error_msg)
            nonlocal done, retry, api_error
            retry +=1
            if retry >= self.max_retries:
                done = True
                api_error = error_msg

        while not done:
            try:
                if data:
                    post_data = { "id": ",".join(data) } 
                    response = requests.post(url_str, data=post_data, timeout=self.timeout)
                else:
                    response = requests.get(url_str, timeout=self.timeout)

                response.raise_for_status()  # Check for HTTP errors
                if response.status_code == 200:
                    done = True
                    api_result = response.text
                elif response.status_code == 429:
                    handle_error(f"Request limiter hit. waiting 2 seconds [Retry: {retry + 1}] code: {response.status_code}")
                    time.sleep(2)
                else:
                    handle_error(f"Failed to retrieve data. | Retry- {retry +1} | Response code- {response.status_code}")

            except Exception as ex:
                aviod_logging_interpolation=f"Error while calling url_str {str(ex)}"
                logger.error(aviod_logging_interpolation)
                error_msg=f"Unexpected Error | Retry- {retry+1} | Response msg- {str(ex)}"
                if isinstance(ex, requests.exceptions.HTTPError):
                        error_msg = f"Check the format of the http request [Retry: {retry + 1}] code: {str(ex)}"
                elif isinstance(ex, requests.exceptions.ConnectionError):
                        error_msg = f"Connection Error [Retry: {retry + 1}] code: {str(ex)}"
                elif isinstance(ex, requests.exceptions.Timeout):
                        error_msg = f"Timeout Error [Retry: {retry + 1}] code: {str(ex)}"
                handle_error(error_msg)

        if api_result is None:
            api_result = f"<response><rest_api_error>{api_error}</rest_api_error></response>"

        if logger.isEnabledFor(logging.DEBUG):
            soup = BeautifulSoup(api_result, "xml")
            # Pretty-print the XML content
            pretty_data = soup.prettify()
            milliseconds = int(round(time.time() * 1000))
            timestamp_str = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + f".{milliseconds % 1000:03d}"
            #with open(f"http_response-{timestamp_str}.xml", 'w') as file:
            #    file.write(pretty_data)
            logger.debug(pretty_data)

        return api_result


    def entreze_esearch(self, params):
        self.function="esearch"
        esearch_params = {}
        esearch_params['db']         = params.get('db', 'pubmed')
        esearch_params['retmax']     = params.get('retmax', '200')
        esearch_params['usehistory'] = "y"
        esearch_params['term']       = params.get('term', None)
        if esearch_params['term'] is None:
            logger.debug("Param 'searchTerm' is required but not passed.")

        api_result = self._rest_api_call(esearch_params)
        ret_params = self._history_key_to_json(api_result)
        return ret_params

    def entreze_epost(self, data, params={}):
        self.function="epost"
        epost_params = {}
        epost_params['db']     = params.get('db', 'pubmed')
        if data is None:
            logger.debug("Param 'data' is required but not passed.")

        api_result = self._rest_api_call(epost_params, data)
        ret_params = self._history_key_to_json(api_result)
        ret_params['count'] = len(data)
        ret_params['db'] =  epost_params['db']
        return ret_params

    def entreze_pmid_summaries(self, params):
        paper_summarys = []
        rec_count = int(params.get('count', 0))
        restart = 0
        while rec_count > 0 :
            params['retstart'] = str(restart)
            params['retmax']  = '200'
            
            soup = self._entreze_get_data(params, "esummary")
            if soup is None:
                return paper_summarys
            
            # Extract information for each UID
            for doc in soup.find_all("DocSum"):
                uid         = self._get_tag_text(doc, "Id")
                issn        = self._get_tag_text(doc, "Item", {"Name": "ISSN"})
                essn        = self._get_tag_text(doc, "Item", {"Name": "ESSN"})
                last_author = self._get_tag_text(doc, "Item", {"Name": "LastAuthor"})
                pmc_id      = self._get_tag_text(doc, "Item", {"Name": "pmc"})
                title       = self._get_tag_text(doc, "Item", {"Name": "Title"})
                source      = self._get_tag_text(doc, "Item", {"Name": "Source"})
                
                paper_summary = {
                    "uid"         : uid,
                    "issn"        : issn,
                    "essn"        : essn,
                    "last_author" : last_author,
                    "pmc_id"      : pmc_id,
                    "title"       : title,
                    "source"      : source
                }
                paper_summarys.append(paper_summary)
                
            restart   +=200 # Increment record position by 200
            rec_count -=200 # Pull the next 200 records or however many are remaining

        return paper_summarys

    def entreze_elink_pmid_to_pmcid(self, params):
        logger.debug("Entering entreze_elink_pmid_to_pmcid!!")
        self.function="elink"
        elink_results = []

        rec_count = int(params.get('count', 0))
        restart = 0
        entrez_params = {}
        entrez_params['dbfrom'] = 'pubmed'
        entrez_params['dbto'] = 'pmc'
        entrez_params['linkname'] = 'pubmed_pmc'

        required_params = ['query_key', 'WebEnv']
        for param_name in required_params:
            if param_name in params:
                entrez_params[param_name] = params[param_name]
            else:
                logger.debug(f"Param '{param_name}' is required but not passed")
                return None
            
        logger.debug(f"XXXXXXXX entreze_elink_pmid_to_pmcid '{rec_count}'")
        while rec_count > 0 :
            params['restart'] = restart

            api_result = self._rest_api_call(entrez_params)
            soup = BeautifulSoup(api_result, "xml")

            api_error = self._get_tag_text(soup,"rest_api_error")
            if api_error:
                logger.error("Error in entreze_elink_pmid_to_pmcid")
                # Maybe throw and exception here
                return None

            root_element = soup.find()
            if root_element.name == 'eLinkResult':
                logger.debug("root_element == eLinkResult!!")

                link_set_db = self._get_tag(soup, ['LinkSetDb'])
                if link_set_db:
                    id_elements = link_set_db.find_all('Id')
                    id_numbers = [id_element.get_text(strip=True) for id_element in id_elements]
                    # Get PubmedArticleSet
                    elink_results += id_numbers


            restart   +=200 # Increment record position by 200
            rec_count -=200 # Pull the next 200 records or however many are remaining
        
        return elink_results

    def entreze_efetch(self, params):
        logger.debug("Entering entreze_efetch!!")
        efetch_results = {'articles':[], 'references':[], 'authors':[]}

        rec_count = int(params.get('count', 0))
        restart = 0
        while rec_count > 0 :
            params['retstart'] = str(restart)
            params['retmax']  = '200'
            logger.debug(f"ZYYYYY {params} ")
            soup = self._entreze_get_data(params, "efetch")
            if soup is None:
                return efetch_results
            
            root_element = soup.find()
            if root_element.name == 'PubmedArticleSet':
                logger.debug("root_element == PubmedArticleSet!!")
                efetch_results['articles']   += self._get_pubmed_articles(soup)
                efetch_results['references'] += self._get_pubmed_references(soup)
                efetch_results['authors']    += self._get_pubmed_authors(soup)
            elif root_element.name == 'pmc-articleset':
                logger.debug("root_element == pmc-article set!!")
                pubmed_articles = self._get_pmc_articles(soup)
                # Get PubmedArticleSet
                efetch_results['articles'] += pubmed_articles

            restart   +=200 # Increment record position by 200
            rec_count -=200 # Pull the next 200 records or however many are remaining
        
        return efetch_results
    
    ############# entrez/eutils Helper funtions #############
    
    def _history_key_to_json(self, api_result):
        ret_params = {}
        ##ret_params['function'] = self.function
        # Parse the XML response using BeautifulSoup
        soup = BeautifulSoup(api_result, "xml")
        # Extract WebEnv and QueryKey
        count     = self._get_tag_text(soup, "Count")
        ret_max   = self._get_tag_text(soup, "RetMax")
        ret_start = self._get_tag_text(soup, "RetStart")
        web_env   = self._get_tag_text(soup, "WebEnv")
        query_key = self._get_tag_text(soup, "QueryKey")
        api_error = self._get_tag_text(soup, "rest_api_error")
        if count:
            ret_params['count'] = count
        if ret_max:
            ret_params['retmax'] = int(ret_max)
        if ret_start:
            ret_params['retstart'] = int(ret_start)

        ret_params['query_key'] = query_key
        ret_params['WebEnv']    = web_env
        if api_error:
            logger.error("Error in history_key_to_json")
            ret_params = {}
            ret_params["Error"] = api_error
        return ret_params


    def _get_pubmed_articles(self, soup):
        articles = []

        pubmed_articles = soup.find_all('PubmedArticle')
        # Iterate over the <PubmedArticle> elements
        for pubmed_article in pubmed_articles:
            logger.debug(f"_get_pubmed_articles  {pubmed_article}")
            article = {}
            medline_citation = self._get_tag(pubmed_article, ['MedlineCitation'])
            article_details  = self._get_tag(medline_citation, ['Article'])
            abstract_details = self._get_tag(article_details, ['Abstract'])
            journal          = self._get_tag(article_details, ['Journal'])
            pub_date         = self._get_tag(journal, ['JournalIssue', 'PubDate'])
            article_id_list  = self._get_tag(pubmed_article, ['PubmedData', 'ArticleIdList'])

            article['pmid']      = self._get_tag_text(medline_citation, "PMID")
            article['issn']      = self._get_tag_text(journal, "ISSN", {"IssnType": "Print"})
            article['eissn']     = self._get_tag_text(journal, "ISSN", {"IssnType": "Electronic"})
            article['pmc']       = self._get_tag_text(article_id_list, "ArticleId", {"IdType": "pmc"})
            article['doi']       = self._get_tag_text(article_id_list, "ArticleId", {"IdType": "doi"})
            article['pub_year']  = self._get_tag_text(pub_date, "Year")
            article['pub_month'] = self._get_tag_text(pub_date, "Month")
            article['pub_day']   = self._get_tag_text(pub_date, "Day")

            article['pub_abbr'] = self._get_tag_text(journal, "ISOAbbreviation")
            article['title']    = self._get_tag_text(article_details, "ArticleTitle")

            abstract_text = ""
            if abstract_details:
                abstract_text = abstract_details.get_text(separator=' ', strip=True)
            article['abstract']  = self._clean_data(abstract_text)

            if article['issn']:
                article['impact_factor'] = get_impact_factor(article['issn'])
            if 'impact_factor' not in article:
                article['impact_factor'] = get_impact_factor(article['eissn'])

            articles.append(article)

        return articles

    def _get_pubmed_authors(self, soup):
            authors = []

            pubmed_articles = soup.find_all('PubmedArticle')
            # Iterate over the <PubmedArticle> elements
            for pubmed_article in pubmed_articles:
                medline_citation = self._get_tag(pubmed_article, ['MedlineCitation'])

                source_pmid        = self._get_tag_text(medline_citation, "PMID")
                author_list     = self._get_tag(pubmed_article, ['AuthorList'])
                if author_list:
                    article_authors = author_list.find_all('Author')
                    for article_author in article_authors:
                        author = {'source_pmid':source_pmid}
                        author['last_name']  = self._get_tag_text(article_author, "LastName")
                        author['first_name'] = self._get_tag_text(article_author, "ForeName")
                        author['initials']   = self._get_tag_text(article_author, "Initials")
                        author['orcid']      = self._get_tag_text(article_author, "Identifier", {"Source": "ORCID"})

                        affiliation_info = self._get_tag(article_author, ['AffiliationInfo'])
                        author['affiliation'] = ''
                        if affiliation_info:
                            author['affiliation'] = self._get_tag_text(affiliation_info, "Affiliation")

                        authors.append(author)

            return authors

    def _get_pubmed_references(self, soup):
            references = []

            pubmed_articles = soup.find_all('PubmedArticle')
            # Iterate over the <PubmedArticle> elements
            for pubmed_article in pubmed_articles:
                medline_citation = self._get_tag(pubmed_article, ['MedlineCitation'])

                source_pmid        = self._get_tag_text(medline_citation, "PMID")
                reference_list     = self._get_tag(pubmed_article, ['ReferenceList'])
                if reference_list:
                    article_references = reference_list.find_all('Reference')
                    for article_reference in article_references:
                        reference = {'source_pmid':source_pmid}
                        reference['citation'] = self._get_tag_text(article_reference, "Citation")
                        article_id_list       = self._get_tag(article_reference, ['ArticleIdList'])
                        reference['pmid']     = self._get_tag_text(article_id_list, "ArticleId", {"IdType": "pubmed"})
                        reference['pmc']      = self._get_tag_text(article_id_list, "ArticleId", {"IdType": "pmc"})
                        references.append(reference)

            return references

    def _clean_data(self, data):
            data = data.encode("ascii", "ignore").decode()
            data = data.replace('\n',' ')
            data = data.replace('"','')
            data = data.replace("\x84", " ")
            return data
    
    def _get_pmc_articles(self, soup):
        
        articles = []
        pmc_articles = soup.find_all('article')
        for pmc_article in pmc_articles:
            #logger.debug(f"_get_pmc_articles  {pmc_article}")
            article = {}
            article['pmid']      = self._get_tag_text(pmc_article, 'article-id', {'pub-id-type': 'pmid'})
            article['pmcid']     = self._get_tag_text(pmc_article, 'article-id', {'pub-id-type': 'pmc'})
            article['publisher'] = self._clean_data(self._get_tag_text(pmc_article, "publisher-name"))
            article['title']      = self._clean_data(self._get_tag_text(pmc_article, "article-title"))
            article['abstract']  = self._clean_data(self._get_tag_text(pmc_article, "abstract"))
            
            body_tag = pmc_article.find('body')
            body_text = ""
            if body_tag:
                body_text = body_tag.get_text(separator=' ', strip=True)
            article['body']  = self._clean_data(body_text)
          
            articles.append(article)

        return articles



    def _entreze_get_data(self, params, function):
        """
        Helper function to call _rest_api_call

        Args:
        params: The parameter to be used to call entrez/eutils.
        function (str): The entrez function to call ['efetch' or 'esummary'].

        Returns:
            soup: Beautiful Soup Response XML

        Note: Both efetch and esummary use the same setup code before calling _rest_api_call
        """
        self.function = function
        entrez_params = {}
        entrez_params['db']  = params.get('db', 'pubmed')
        entrez_params['retmax']  = params.get('retmax', '200')
        entrez_params['retstart']  = params.get('retstart', '0')


        required_params = ['query_key', 'WebEnv']
        for param_name in required_params:
            if param_name in params:
                entrez_params[param_name] = params[param_name]
            else:
                logger.debug(f"Param '{param_name}' is required but not passed")
                return None

        api_result = self._rest_api_call(entrez_params)
        # Parse the XML response using BeautifulSoup
        soup = BeautifulSoup(api_result, "xml")

        api_error = self._get_tag_text(soup,"rest_api_error")
        if api_error:
            logger.error("Error in _entreze_get_data")
            # Maybe throw and exception here
            return None
        
        return soup
    
    ############# Beautiful Soup XML Helper funtions #############

    def _get_tag(self, soup, path_names):
        root = soup
        for path_name in path_names:
            #logger.debug(f"get_tag {type(root)} {root} {path_name}")
            root = root.find(path_name, default='')
        return root

    def _get_tag_text(self, soup, tag_name, attribute=None):
        ret_val = ""
        if soup is not None:
            try:
                if attribute:
                    tag = soup.find(tag_name, attribute)
                else:
                    tag = soup.find(tag_name)

                ret_val = tag.text if tag else ""
            except Exception as e:
                print(f"Error finding tag in _get_tag_text: {e}")
                ret_val = ""
        else:
            ret_val = ""

        return ret_val




import pytest
import json
import inspect
import os
from pub_worm.orcid.orcid_api import OrcidAPI

DUMP_API_CALL = True
def dump_api_call(function_name, actual_result):
    if DUMP_API_CALL:
        pretty_data = json.dumps(actual_result, indent=4)
        output_path = "output"
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        with open(f"{output_path}/{function_name}.json", 'w') as file:
                file.write(pretty_data)
        print(pretty_data)

def test_get_orcid_data_from_orcid_id():
    function_name = inspect.currentframe().f_code.co_name
    orcid_id = "0000-0002-2403-8551"
    orcid_api = OrcidAPI()
    actual_result = orcid_api.get_orcid_data(orcid_id)
    dump_api_call(function_name, actual_result)
    
    expected_result = "Higgins"
    assert actual_result['person']['name']['family-name']['value'] == expected_result

if __name__ == "__main__":
    pytest.main([__file__])
